/*
 *  Copyright (c) 2004-2020 Mellanox Technologies LTD. All rights reserved
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include "infiniband/ibdiag/ibdiag.h"
#include "infiniband/ibdiag/ibdiag_clbck.h"

extern IBDiagClbck ibDiagClbck;

int IBDiag::BuildEnhancedCC(list_p_fabric_general_err& congestion_control_errors,
                            progress_func_nodes_t progress_func)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    bool all_devices_sup_cc = true;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ibDiagClbck.Set(this, &fabric_extended_info, &congestion_control_errors);
    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    CC_EnhancedCongestionInfo cc_congestion_info;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func = IBDiagCCEnhancedCongestionInfoGetClbck;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                    " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes, &this->discover_progress_bar_nodes);

        // check that Enhanced Congestion Info supported
        if (!this->capability_module.IsSupportedSMPCapability(p_curr_node,
                EnSMPCapIsVSCCSupported)) {
            all_devices_sup_cc = false;
            continue;
        }

        IBPort *p_curr_port = NULL;
        if (p_curr_node->type == IB_SW_NODE) {
            p_curr_port = p_curr_node->getPort(0);
        } else {
            for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {
                p_curr_port = p_curr_node->getPort((phys_port_t)i);
                if (p_curr_port && p_curr_port->is_data_worthy())
                    break;
            }
        }
        if (!p_curr_port)
            continue;

        clbck_data.m_data1 = p_curr_node;
        uint16_t lid = p_curr_port->base_lid;

        // 1.1 Enhanced Congestion Info
        this->ibis_obj.CCEnhancedInfoGet(lid,
                                         0, // sl
                                         0, // cc key
                                         &cc_congestion_info,
                                         &clbck_data);

        if(ibDiagClbck.GetState())
           goto exit;
    } // nodes loop

    // set a warning if some devices not support CC
    if (!all_devices_sup_cc) {

        FabricErrNotAllDevicesSupCap *p_curr_fabric_err =
                new FabricErrNotAllDevicesSupCap("Enhanced Congestion Control");

        if (!p_curr_fabric_err) {
            this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
            rc = IBDIAG_ERR_CODE_NO_MEM;
            goto exit;
        }

        congestion_control_errors.push_back(p_curr_fabric_err);
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!congestion_control_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::BuildCCSwithConfig(list_p_fabric_general_err& congestion_control_errors,
                               progress_func_nodes_t progress_func)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    CC_CongestionSwitchGeneralSettings curr_switch_general_settings;
    CC_CongestionPortProfileSettings curr_port_profile_settings;
    CC_CongestionSLMappingSettings curr_sl_mapping_settings;

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    ibDiagClbck.Set(this, &fabric_extended_info, &congestion_control_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                    " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        // this MADs are only for switches
        if (p_curr_node->type != IB_SW_NODE)
            continue;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes, &this->discover_progress_bar_nodes);

        // check if Enhanced Congestion Control is supported
        CC_EnhancedCongestionInfo *p_cc_enhanced_info =
                fabric_extended_info.getCCEnhancedCongestionInfo(p_curr_node->createIndex);
        if (!p_cc_enhanced_info) // not supported
            continue;
        if (!p_cc_enhanced_info->ver0) { // no version supported
            FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                    new FabricErrNodeNotSupportCap(p_curr_node,
                            "This device does not support " \
                            "any version of Congestion Control attributes");

            if (!p_curr_fabric_node_err) {
                this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
                rc = IBDIAG_ERR_CODE_NO_MEM;
                goto exit;
            }

            congestion_control_errors.push_back(p_curr_fabric_node_err);
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            continue;
        }

        IBPort *p_curr_port0 = p_curr_node->getPort(0); // get switch port 0
        if (!p_curr_port0)
            continue;

        uint16_t lid = p_curr_port0->base_lid;

        // 1.2 Congestion Switch General Settings
        clbck_data.m_data1 = p_curr_node;
        clbck_data.m_handle_data_func = IBDiagCCSwitchGeneralSettingsGetClbck;
        this->ibis_obj.CCSwitchGeneralSettingsGet(lid,
                                                  0, // sl
                                                  0, // cc key
                                                  &curr_switch_general_settings,
                                                  &clbck_data);
        if(ibDiagClbck.GetState())
            goto exit;

        // go over all ports of this switch:
        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_sw_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_sw_port || !p_curr_sw_port->is_data_worthy())
                continue;

            // skip ANs' ports
            if (GetSpecialCAPortType(p_curr_sw_port->p_remotePort->p_node) != NotSpecial)
                continue;

            clbck_data.m_data1 = p_curr_sw_port;

            // 1.3 Congestion Port Profile Settings
            clbck_data.m_handle_data_func = IBDiagCCPortProfileSettingsGetClbck;
            // get OperationalVLs for this port
            u_int8_t op_vls = fabric_extended_info.getSMPPortInfo(p_curr_sw_port->createIndex)->OpVLs;
            u_int8_t vl_num = get_operational_vl_num(op_vls);

            // go over all OperationalVLs of this port
            for (u_int8_t vl = 0; vl < vl_num; ++vl) {
                CLEAR_STRUCT(curr_port_profile_settings);
                curr_port_profile_settings.vl_mask = (u_int16_t)(1 << vl);
                clbck_data.m_data2 = (void *)(uintptr_t)vl;

                this->ibis_obj.CCPortProfileSettingsGet(lid,
                                                        pi, // port
                                                        0,  // sl
                                                        0,  // cc key
                                                        &curr_port_profile_settings,
                                                        &clbck_data);
                if(ibDiagClbck.GetState())
                    goto exit;
            }

            // 1.4 Congestion SL Mapping Settings
            clbck_data.m_handle_data_func = IBDiagCCSLMappingSettingsGetClbck;

            this->ibis_obj.CCSLMappingSettingsGet(lid,
                                                  pi, // port
                                                  0,  // sl
                                                  0,  // cc key
                                                  &curr_sl_mapping_settings,
                                                  &clbck_data);
            if(ibDiagClbck.GetState())
                goto exit;

        } // ports loop
    } // nodes loop

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!congestion_control_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


int IBDiag::BuildCCHCAConfig(list_p_fabric_general_err& congestion_control_errors,
                               progress_func_nodes_t progress_func)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    CC_CongestionHCAGeneralSettings curr_hca_general_settings;
    CC_CongestionHCARPParameters    curr_hca_rp_parameters;
    CC_CongestionHCANPParameters    curr_hca_np_parameters;

    progress_bar_nodes_t progress_bar_nodes;
    CLEAR_STRUCT(progress_bar_nodes);

    ibDiagClbck.Set(this, &fabric_extended_info, &congestion_control_errors);
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                    " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        // this MADs are only for HCAs
        if (p_curr_node->type != IB_CA_NODE)
            continue;

        ++progress_bar_nodes.nodes_found;
        if (p_curr_node->type == IB_SW_NODE)
            ++progress_bar_nodes.sw_found;
        else
            ++progress_bar_nodes.ca_found;
        if (progress_func)
            progress_func(&progress_bar_nodes, &this->discover_progress_bar_nodes);

        // check if Enhanced Congestion Control is supported
        CC_EnhancedCongestionInfo *p_cc_enhanced_info =
                fabric_extended_info.getCCEnhancedCongestionInfo(p_curr_node->createIndex);
        if (!p_cc_enhanced_info) // not supported
            continue;
        if (!p_cc_enhanced_info->ver0) { // no version supported
            FabricErrNodeNotSupportCap *p_curr_fabric_node_err =
                    new FabricErrNodeNotSupportCap(p_curr_node,
                            "This device does not support " \
                            "any version of Congestion Control attributes");

            if (!p_curr_fabric_node_err) {
                this->SetLastError("Failed to allocate FabricErrNodeNotSupportCap");
                rc = IBDIAG_ERR_CODE_NO_MEM;
                goto exit;
            }

            congestion_control_errors.push_back(p_curr_fabric_node_err);
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            continue;
        }

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            clbck_data.m_data1 = p_curr_port;
            uint16_t lid = p_curr_port->base_lid;

            // 1.5 Congestion HCA General Settings
            clbck_data.m_handle_data_func = IBDiagCCHCAGeneralSettingsGetClbck;
            this->ibis_obj.CCHCAGeneralSettingsGet(lid,
                                                   0, // sl
                                                   0, // cc key
                                                   &curr_hca_general_settings,
                                                   &clbck_data);
            if(ibDiagClbck.GetState())
                goto exit;

            // 1.6 Congestion HCA RP Parameters
            clbck_data.m_handle_data_func = IBDiagCCHCARPParametersGetClbck;
            this->ibis_obj.CCHCARPParametersGet(lid,
                                                0, // sl
                                                0, // cc key
                                                &curr_hca_rp_parameters,
                                                &clbck_data);
            if(ibDiagClbck.GetState())
                goto exit;

            // 1.7 Congestion HCA NP Parameters
            clbck_data.m_handle_data_func = IBDiagCCHCANPParametersGetClbck;
            this->ibis_obj.CCHCANPParametersGet(lid,
                                                0, // sl
                                                0, // cc key
                                                &curr_hca_np_parameters,
                                                &clbck_data);
            if(ibDiagClbck.GetState())
                goto exit;

        } // ports loop
    } // nodes loop

exit:
    this->ibis_obj.MadRecAll();

    if (rc == IBDIAG_SUCCESS_CODE) {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!congestion_control_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}


void IBDiag::DumpCCEnhancedInfoCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    csv_out.DumpStart(SECTION_CC_ENHANCED_INFO);

    stringstream sstream;
    sstream << "NodeGUID," << "ver0Supported" << endl;
    csv_out.WriteLine(sstream.str());

    struct CC_EnhancedCongestionInfo *p_curr_cc_enhanced_info = NULL;
    char curr_line[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node)                   //node not exist
            continue;

        p_curr_cc_enhanced_info = this->fabric_extended_info.
                getCCEnhancedCongestionInfo(p_curr_node->createIndex);
        if (!p_curr_cc_enhanced_info)
            continue;

        memset(curr_line, 0, sizeof(curr_line));
        sstream.str("");

        sprintf(curr_line, U64H_FMT ",%u",
                p_curr_node->guid_get(),
                p_curr_cc_enhanced_info->ver0);

        sstream << curr_line << endl;
        csv_out.WriteLine(sstream.str());
    }

    csv_out.DumpEnd(SECTION_CC_ENHANCED_INFO);
    IBDIAG_RETURN_VOID;
}


void IBDiag::DumpCCSwitchGeneralSettingsCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    csv_out.DumpStart(SECTION_CC_SWITCH_GENERAL_SETTINGS);

    stringstream sstream;
    sstream << "NodeGUID," << "aqs_time," << "aqs_weight,"
            << "en," << "cap_total_buffer_size" << endl;
    csv_out.WriteLine(sstream.str());

    struct CC_CongestionSwitchGeneralSettings *p_curr_cc_switch_general_settings = NULL;
    char curr_line[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node)                   //node not exist
            continue;

        p_curr_cc_switch_general_settings = this->fabric_extended_info.
                getCCSwitchGeneralSettings(p_curr_node->createIndex);
        if (!p_curr_cc_switch_general_settings)
            continue;

        memset(curr_line, 0, sizeof(curr_line));
        sstream.str("");

        sprintf(curr_line, U64H_FMT ",%u,%u,%u,%u",
                p_curr_node->guid_get(),
                p_curr_cc_switch_general_settings->aqs_time,
                p_curr_cc_switch_general_settings->aqs_weight,
                p_curr_cc_switch_general_settings->en,
                p_curr_cc_switch_general_settings->cap_total_buffer_size);

        sstream << curr_line << endl;
        csv_out.WriteLine(sstream.str());
    }

    csv_out.DumpEnd(SECTION_CC_SWITCH_GENERAL_SETTINGS);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpCCPortProfileSettingsCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    csv_out.DumpStart(SECTION_CC_PORT_PROFILE_SETTINGS);

    stringstream sstream;
    sstream << "NodeGUID," << "portNum," << "vl," << "mode,"
            << "profile1_min," << "profile1_max," << "profile1_percent,"
            << "profile2_min," << "profile2_max," << "profile2_percent,"
            << "profile3_min," << "profile3_max," << "profile3_percent"
            << endl;

    csv_out.WriteLine(sstream.str());
    struct CC_CongestionPortProfileSettings *p_curr_cc_port_profile_settings = NULL;
    char curr_line[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node)                   // node not exist
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            SMP_PortInfo *p_curr_port_info = fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
            if (!p_curr_port_info)          // port info not exist
                continue;

            u_int8_t vl_num = get_operational_vl_num(p_curr_port_info->OpVLs);
            // go over all OperationalVLs of this port
            for (u_int8_t vl = 0; vl < vl_num; ++vl) {

                p_curr_cc_port_profile_settings = this->fabric_extended_info.
                        getCCPortProfileSettings(p_curr_port->createIndex, vl);
                if (!p_curr_cc_port_profile_settings)
                    continue;

                memset(curr_line, 0, sizeof(curr_line));
                sstream.str("");

                sprintf(curr_line, U64H_FMT ",%u,%u,%u,"
                        "%u,%u,%u,%u,%u,%u,%u,%u,%u",
                        p_curr_node->guid_get(),
                        p_curr_port->num,
                        vl,
                        p_curr_cc_port_profile_settings->mode,
                        p_curr_cc_port_profile_settings->profile1_min,
                        p_curr_cc_port_profile_settings->profile1_max,
                        p_curr_cc_port_profile_settings->profile1_percent,
                        p_curr_cc_port_profile_settings->profile2_min,
                        p_curr_cc_port_profile_settings->profile2_max,
                        p_curr_cc_port_profile_settings->profile2_percent,
                        p_curr_cc_port_profile_settings->profile3_min,
                        p_curr_cc_port_profile_settings->profile3_max,
                        p_curr_cc_port_profile_settings->profile3_percent);

                sstream << curr_line << endl;
                csv_out.WriteLine(sstream.str());

            } // VLs
        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_PORT_PROFILE_SETTINGS);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpCCSLMappingSettingsCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    csv_out.DumpStart(SECTION_CC_SL_MAPPING_SETTINGS);

    stringstream sstream;
    sstream << "NodeGUID," << "portNum,"
            << "sl_profile_0," << "sl_profile_1," << "sl_profile_2," << "sl_profile_3,"
            << "sl_profile_4," << "sl_profile_5," << "sl_profile_6," << "sl_profile_7,"
            << "sl_profile_8," << "sl_profile_9," << "sl_profile_10," << "sl_profile_11,"
            << "sl_profile_12," << "sl_profile_13," << "sl_profile_14," << "sl_profile_15"
            << endl;

    csv_out.WriteLine(sstream.str());
    struct CC_CongestionSLMappingSettings *p_curr_cc_sl_mapping_settings = NULL;
    char curr_line[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node)                   // node not exist
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            p_curr_cc_sl_mapping_settings = this->fabric_extended_info.
                    getCCSLMappingSettings(p_curr_port->createIndex);
            if (!p_curr_cc_sl_mapping_settings)
                continue;

            memset(curr_line, 0, sizeof(curr_line));
            sstream.str("");

            sprintf(curr_line, U64H_FMT ",%u,"
                    "%u,%u,%u,%u,"
                    "%u,%u,%u,%u,"
                    "%u,%u,%u,%u,"
                    "%u,%u,%u,%u",
                    p_curr_node->guid_get(),
                    p_curr_port->num,
                    p_curr_cc_sl_mapping_settings->sl_profile_0,
                    p_curr_cc_sl_mapping_settings->sl_profile_1,
                    p_curr_cc_sl_mapping_settings->sl_profile_2,
                    p_curr_cc_sl_mapping_settings->sl_profile_3,
                    p_curr_cc_sl_mapping_settings->sl_profile_4,
                    p_curr_cc_sl_mapping_settings->sl_profile_5,
                    p_curr_cc_sl_mapping_settings->sl_profile_6,
                    p_curr_cc_sl_mapping_settings->sl_profile_7,
                    p_curr_cc_sl_mapping_settings->sl_profile_8,
                    p_curr_cc_sl_mapping_settings->sl_profile_9,
                    p_curr_cc_sl_mapping_settings->sl_profile_10,
                    p_curr_cc_sl_mapping_settings->sl_profile_11,
                    p_curr_cc_sl_mapping_settings->sl_profile_12,
                    p_curr_cc_sl_mapping_settings->sl_profile_13,
                    p_curr_cc_sl_mapping_settings->sl_profile_14,
                    p_curr_cc_sl_mapping_settings->sl_profile_15);

            sstream << curr_line << endl;
            csv_out.WriteLine(sstream.str());

        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_SL_MAPPING_SETTINGS);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpCCHCAGeneralSettingsCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    csv_out.DumpStart(SECTION_CC_HCA_GENERAL_SETTINGS);

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," << "portNum," << "en_react," << "en_notify"
            << endl;

    csv_out.WriteLine(sstream.str());
    struct CC_CongestionHCAGeneralSettings *p_curr_cc_hca_general_settings = NULL;
    char curr_line[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
         IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
         if (!p_curr_node)                   // node not exist
             continue;

         for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
             IBPort *p_curr_port = p_curr_node->getPort(pi);
             //Skip ports that don't exist or are down or have undefined state
             //or are not in the scope of the fabric.
             if (!p_curr_port || !p_curr_port->is_data_worthy())
                 continue;

             p_curr_cc_hca_general_settings = this->fabric_extended_info.
                     getCCHCAGeneralSettings(p_curr_port->createIndex);
             if (!p_curr_cc_hca_general_settings)
                 continue;

             memset(curr_line, 0, sizeof(curr_line));
             sstream.str("");

             sprintf(curr_line, U64H_FMT "," U64H_FMT ",%u,%u,%u",
                     p_curr_node->guid_get(),
                     p_curr_port->guid_get(),
                     p_curr_port->num,
                     p_curr_cc_hca_general_settings->en_react,
                     p_curr_cc_hca_general_settings->en_notify);

             sstream << curr_line << endl;
             csv_out.WriteLine(sstream.str());

         } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_HCA_GENERAL_SETTINGS);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpCCHCARPParametersCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    csv_out.DumpStart(SECTION_CC_HCA_RP_PARAMETERS);

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," << "portNum," << "clamp_tgt_rate_after_time_inc,"
            << "clamp_tgt_rate," << "rpg_time_reset," << "rpg_byte_reset," << "rpg_threshold,"
            << "rpg_max_rate," << "rpg_ai_rate," << "rpg_hai_rate," << "rpg_gd," << "rpg_min_dec_fac,"
            << "rpg_min_rate," << "rate_to_set_on_first_cnp," << "dce_tcp_g," << "dce_tcp_rtt,"
            << "rate_reduce_mionitor_period," << "initial_alpha_value"
            << endl;

    csv_out.WriteLine(sstream.str());
    struct CC_CongestionHCARPParameters *p_curr_cc_hca_rp_parameters = NULL;
    char curr_line[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node)                   // node not exist
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            p_curr_cc_hca_rp_parameters = this->fabric_extended_info.
                    getCCHCARPParameters(p_curr_port->createIndex);
            if (!p_curr_cc_hca_rp_parameters)
                continue;

            memset(curr_line, 0, sizeof(curr_line));
            sstream.str("");

            sprintf(curr_line, U64H_FMT "," U64H_FMT ",%u,"
                    "%u,%u,%u,%u,"
                    "%u,%u,%u,%u,"
                    "%u,%u,%u,%u,"
                    "%u,%u,%u,%u",
                    p_curr_node->guid_get(),
                    p_curr_port->guid_get(),
                    p_curr_port->num,
                    p_curr_cc_hca_rp_parameters->clamp_tgt_rate_after_time_inc,
                    p_curr_cc_hca_rp_parameters->clamp_tgt_rate,
                    p_curr_cc_hca_rp_parameters->rpg_time_reset,
                    p_curr_cc_hca_rp_parameters->rpg_byte_reset,
                    p_curr_cc_hca_rp_parameters->rpg_threshold,
                    p_curr_cc_hca_rp_parameters->rpg_max_rate,
                    p_curr_cc_hca_rp_parameters->rpg_ai_rate,
                    p_curr_cc_hca_rp_parameters->rpg_hai_rate,
                    p_curr_cc_hca_rp_parameters->rpg_gd,
                    p_curr_cc_hca_rp_parameters->rpg_min_dec_fac,
                    p_curr_cc_hca_rp_parameters->rpg_min_rate,
                    p_curr_cc_hca_rp_parameters->rate_to_set_on_first_cnp,
                    p_curr_cc_hca_rp_parameters->dce_tcp_g,
                    p_curr_cc_hca_rp_parameters->dce_tcp_rtt,
                    p_curr_cc_hca_rp_parameters->rate_reduce_mionitor_period,
                    p_curr_cc_hca_rp_parameters->initial_alpha_value);

            sstream << curr_line << endl;
            csv_out.WriteLine(sstream.str());

        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_HCA_RP_PARAMETERS);
    IBDIAG_RETURN_VOID;
}

void IBDiag::DumpCCHCANPParametersCSVTable(CSVOut &csv_out)
{
    IBDIAG_ENTER;
    csv_out.DumpStart(SECTION_CC_HCA_NP_PARAMETERS);

    stringstream sstream;
    sstream << "NodeGUID," << "PortGUID," << "portNum," << "min_time_between_cnps,"
            << "cnp_sl," <<"cnp_sl_mode"
            << endl;

    csv_out.WriteLine(sstream.str());
    struct CC_CongestionHCANPParameters *p_curr_cc_hca_np_parameters = NULL;
    char curr_line[1024];

    for (u_int32_t nI = 0; nI < this->fabric_extended_info.getNodesVectorSize(); ++nI) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(nI);
        if (!p_curr_node)                   // node not exist
            continue;

        for (phys_port_t pi = 1; pi < p_curr_node->numPorts + 1; ++pi) {
            IBPort *p_curr_port = p_curr_node->getPort(pi);
            //Skip ports that don't exist or are down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            p_curr_cc_hca_np_parameters = this->fabric_extended_info.
                    getCCHCANPParameters(p_curr_port->createIndex);
            if (!p_curr_cc_hca_np_parameters)
                continue;

            memset(curr_line, 0, sizeof(curr_line));
            sstream.str("");

            sprintf(curr_line, U64H_FMT "," U64H_FMT ",%u,%u,%u,%u",
                    p_curr_node->guid_get(),
                    p_curr_port->guid_get(),
                    p_curr_port->num,
                    p_curr_cc_hca_np_parameters->min_time_between_cnps,
                    p_curr_cc_hca_np_parameters->cnp_sl,
                    p_curr_cc_hca_np_parameters->cnp_sl_mode);

            sstream << curr_line << endl;
            csv_out.WriteLine(sstream.str());

        } // ports
    } // nodes

    csv_out.DumpEnd(SECTION_CC_HCA_NP_PARAMETERS);
    IBDIAG_RETURN_VOID;
}
